/*
    tvrender for libview64
    (c) 2013 Soci/Singular (soci at c64.rulez.org)

    $Id: tvrender.c 42 2015-05-10 18:09:41Z soci $

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

*/

#include <string.h>
#include <stdio.h>
#include <math.h>
#include "tvrender.h"

static rgb_color default_palette[16] = {
    { 0x0f, 0x0f, 0x0f },
    { 0xfa, 0xfa, 0xfa },
    { 0x7f, 0x27, 0x34 },
    { 0x6f, 0xc0, 0xb4 },
    { 0x87, 0x3c, 0x95 },
    { 0x56, 0x9c, 0x49 },
    { 0x3e, 0x3e, 0x8e },
    { 0xbb, 0xbb, 0x6a },
    { 0x86, 0x3d, 0x27 },
    { 0x55, 0x2c, 0x00 },
    { 0xae, 0x56, 0x63 },
    { 0x4e, 0x4e, 0x4e },
    { 0x75, 0x75, 0x75 },
    { 0x94, 0xda, 0x87 },
    { 0x75, 0x75, 0xc5 },
    { 0x9c, 0x9c, 0x9c }
};

static tvrender_image *new_tv_pixels(int w, int h, int p) {
    tvrender_image *pixels;

    pixels = (tvrender_image *)malloc(sizeof(tvrender_image));
    if (pixels == NULL) return NULL;
    pixels->pitch = p * 3;
    pixels->h = h;
    pixels->w = w;
    pixels->d = (unsigned char *)malloc(pixels->pitch * pixels->h);
    if (pixels->d == NULL) {
        free(pixels);
        return NULL;
    }
    return pixels;
}

static void free_tv_pixels(tvrender_image *pixels) {
    free(pixels->d);
    free(pixels);
}

static tvrender_image *create_pal(const view64_screen *rscreen, int scale, int frame, unsigned int vmode) {
    tvrender_image *pixels;
    const static int luma[16] = {
        2, 32, 10, 20, 12, 16, 8, 24, 12, 8, 16, 10, 15, 24, 15, 20
    };
    const static signed char cols[32][16] = {
        { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, // 0
        { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, // 1
        { -24, -4, 16, 35, 48, 53, 51, 40, 24, 4, -16, -35, -48, -53, -51, -40}, // 2
        { 22, 3, -15, -32, -44, -50, -47, -38, -22, -3, 15, 32, 44, 50, 47, 38}, // 3
        { 39, 53, 60, 57, 45, 27, 4, -18, -39, -53, -60, -57, -45, -27, -4, 18}, // 4
        { -36, -50, -56, -53, -42, -25, -4, 17, 36, 50, 56, 53, 42, 25, 4, -17}, // 5
        { 60, 57, 45, 27, 4, -18, -39, -53, -60, -57, -45, -27, -4, 18, 39, 53}, // 6
        { -61, -58, -46, -27, -4, 19, 40, 54, 61, 58, 46, 27, 4, -19, -40, -54}, // 7
        { -44, -26, -4, 18, 38, 52, 58, 56, 44, 26, 4, -18, -38, -52, -58, -56}, // 8
        { -59, -47, -28, -4, 19, 40, 55, 62, 59, 47, 28, 4, -19, -40, -55, -62}, // 9
        { -24, -4, 16, 35, 48, 53, 51, 40, 24, 4, -16, -35, -48, -53, -51, -40}, // 10
        { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, // 11
        { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, // 12
        { -36, -50, -56, -53, -42, -25, -4, 17, 36, 50, 56, 53, 42, 25, 4, -17}, // 13
        { 60, 57, 45, 27, 4, -18, -39, -53, -60, -57, -45, -27, -4, 18, 39, 53}, // 14
        { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, // 15

        { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, // 0
        { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, // 1
        { 16, 35, 48, 53, 51, 40, 24, 4, -16, -35, -48, -53, -51, -40, -24, -4}, // 2
        { -15, -32, -44, -50, -47, -38, -22, -3, 15, 32, 44, 50, 47, 38, 22, 3}, // 3
        { -45, -27, -4, 18, 39, 53, 60, 57, 45, 27, 4, -18, -39, -53, -60, -57}, // 4
        { 42, 25, 4, -17, -36, -50, -56, -53, -42, -25, -4, 17, 36, 50, 56, 53}, // 5
        { -60, -57, -45, -27, -4, 18, 39, 53, 60, 57, 45, 27, 4, -18, -39, -53}, // 6
        { 61, 58, 46, 27, 4, -19, -40, -54, -61, -58, -46, -27, -4, 19, 40, 54}, // 7
        { 38, 52, 58, 56, 44, 26, 4, -18, -38, -52, -58, -56, -44, -26, -4, 18}, // 8
        { 55, 62, 59, 47, 28, 4, -19, -40, -55, -62, -59, -47, -28, -4, 19, 40}, // 9
        { 16, 35, 48, 53, 51, 40, 24, 4, -16, -35, -48, -53, -51, -40, -24, -4}, // 10
        { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, // 11
        { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, // 12
        { 42, 25, 4, -17, -36, -50, -56, -53, -42, -25, -4, 17, 36, 50, 56, 53}, // 13
        { -60, -57, -45, -27, -4, 18, 39, 53, 60, 57, 45, 27, 4, -18, -39, -53}, // 14
        { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, // 15
    };
    unsigned char *p, *d;
    int x, y, w, lb;
    signed char *l, *ol;
    unsigned int *offs = NULL;
    d = rscreen->d[frame];
    pixels = new_tv_pixels(rscreen->w * scale, rscreen->h, (rscreen->w + 2) * scale);
    w = (pixels->pitch * 1000) / scale / 2865;
    lb = ((w * scale - pixels->w) / scale + 1) / 2;

    l = (signed char *)malloc((pixels->pitch + 3 * scale) * 600 / 191 / scale * 9 + 8 + 16);
    if (l == NULL) return NULL;
    ol = (signed char *)malloc((pixels->pitch + 3 * scale) * 600 / 191 / scale * 9 + 8 + 16);
    if (ol == NULL) {
        free(l);
        return NULL;
    }
    for (x = 0; x < (pixels->pitch + 3 * scale) * 600 / 191 / scale * 9 + 8 + 16; x++) ol[x] = cols[rscreen->br + 16][x & 15];
    memcpy(l, ol, (pixels->pitch + 3 * scale) * 600 / 191 / scale * 9 + 8 + 16);

    if (!(vmode & VIDEO_MONOCHROME)) {
        int a, i2, j;
        int scale2 = scale * 8;
        offs = (unsigned int *)malloc(sizeof(unsigned int *) * ((pixels->pitch / 3) + 2 * scale));
        if (offs == NULL) {
            free(ol);
            free(l);
            return NULL;
        }
        for (j = i2 = a = x = 0; x < (pixels->pitch / 3) + 2 * scale; x++) {
            offs[x] = j;
            a += (600 * 3) % 191;
            if (a >= 191) {
                a -= 191;
                i2 += (600 * 3) / 191 + 1;
            } else i2 += (600 * 3) / 191;
            while (i2 >= scale2) {
                i2 -= scale2;
                j += 8;
            }
        }
    }

    memset(pixels->d, 0, pixels->pitch * pixels->h);
    for (y = 0; y < rscreen->h; y++) {
        int i, n = 0, ci = 0;
        int accu = 0;
        signed char *l2;
        p = pixels->d + y * pixels->pitch;
        l2 = l;
        l = ol;
        ol = l2;
        {
            int l;
            const signed char *c;
            int cb = rscreen->b ? rscreen->b[y * 2] : rscreen->br;
            c = cols[cb + 16 * (y & 1)];
            for (l = 0; l < 8; l++) {
                l2[ci] = c[ci];
                ci++;
            }
        }
        for (x = -lb; x < w - lb; x++) {
            int l, ll;
            const signed char *c;
            if (x < 0) {
                int cb = rscreen->b ? rscreen->b[y * 2] : rscreen->br;
                l = luma[cb];
                c = cols[cb + 16 * (y & 1)];
            } else if (x >= rscreen->w) {
                int cb = rscreen->b ? rscreen->b[y * 2 + 1] : rscreen->br;
                l = luma[cb];
                c = cols[cb + 16 * (y & 1)];
            } else {
                l = luma[*d];
                c = cols[*d + 16 * (y & 1)];
                d++;
            }
            accu += (1000 - n) * l;
            *p = accu / 128;
            p++;
            n += 2865 * scale - 1000;
            ll = 1000 * l / 128;
            while (n >= 1000) {
                *p++ = ll;
                n -= 1000;
            }
            accu = n * l;
            if (c[0]) {
                for (l = 0; l < 9; l++) {
                    l2[ci] = c[ci & 15];
                    ci++;
                }
            } else {
                memset(l2 + ci, 0, 9);
                ci += 9;
            }
        }
        p = pixels->d + y * pixels->pitch;
        if (scale == 1) {
            int a = 0, b;
            for (x = 0; x < 5; x++) a += p[x];
            b = p[x];
            p[x] = a;
            for (x = 1; x < pixels->pitch - 4; x++) {
                a += p[x+4] - b;
                b = p[x];
                p[x] = a / 5;
            }
        } else {
            int a = 0, b;
            for (x = 0; x < 7; x++) a += p[x];
            b = p[x];
            p[x] = a;
            for (x = 1; x < pixels->pitch - 6; x++) {
                a += p[x+6] - b;
                b = p[x];
                p[x] = a / 7;
            }
        }
        if (!(vmode & VIDEO_MONOCHROME)) {
            unsigned int *offs1 = offs, *offs2 = offs;
            int r = 0, g = 0, b = 0;
            if (y & 1) {
                for (x = 0; x < 2 * scale; x++) {
                    int j = *offs2++;
                    int e = l[j + 4] + ol[j + 4];
                    if (j & 8) {
                        r -= e;
                        g += e;
                        b += l[j + 8] - ol[j + 8];
                    } else {
                        r += e;
                        g -= e;
                        b -= l[j + 8] - ol[j + 8];
                    }
                }
                for (x = 0; x < pixels->pitch; x += 3) {
                    int j;
                    int v;
                    int j2 = *offs1++;
                    int e = l[j2 + 4] + ol[j2 + 4];
                    if (j2 & 8) {
                        r += e;
                        g -= e;
                        b -= l[j2 + 8] - ol[j2 + 8];
                    } else {
                        r -= e;
                        g += e;
                        b += l[j2 + 8] - ol[j2 + 8];
                    }
                    j = *offs2++;
                    e = l[j + 4] + ol[j + 4];
                    if (j & 8) {
                        r -= e;
                        g += e;
                        b += l[j + 8] - ol[j + 8];
                    } else {
                        r += e;
                        g -= e;
                        b -= l[j + 8] - ol[j + 8];
                    }
                    v = p[x];
                    v += r / 4 / scale;
                    p[x] = v & ~255 ? ((v < 0) ? 0 : 255) : v;
                    v = p[x+1];
                    v += g / 5 / scale;
                    p[x+1] = v & ~255 ? ((v < 0) ? 0 : 255) : v;
                    v = p[x+2];
                    v += b / 3 / scale;
                    p[x+2] = v & ~255 ? ((v < 0) ? 0 : 255) : v;
                }
            } else {
                for (x = 0; x < 2 * scale; x++) {
                    int j = *offs2++;
                    int e = l[j + 4] + ol[j + 4];
                    if (j & 8) {
                        r -= e;
                        g += e;
                        b -= l[j + 8] - ol[j + 8];
                    } else {
                        r += e;
                        g -= e;
                        b += l[j + 8] - ol[j + 8];
                    }
                }
                for (x = 0; x < pixels->pitch; x += 3) {
                    int j;
                    int v;
                    int j2 = *offs1++;
                    int e = l[j2 + 4] + ol[j2 + 4];
                    if (j2 & 8) {
                        r += e;
                        g -= e;
                        b += l[j2 + 8] - ol[j2 + 8];
                    } else {
                        r -= e;
                        g += e;
                        b -= l[j2 + 8] - ol[j2 + 8];
                    }
                    j = *offs2++;
                    e = l[j + 4] + ol[j + 4];
                    if (j & 8) {
                        r -= e;
                        g += e;
                        b -= l[j + 8] - ol[j + 8];
                    } else {
                        r += e;
                        g -= e;
                        b += l[j + 8] - ol[j + 8];
                    }
                    v = p[x];
                    v += r / 4 / scale;
                    p[x] = v & ~255 ? ((v < 0) ? 0 : 255) : v;
                    v = p[x+1];
                    v += g / 5 / scale;
                    p[x+1] = v & ~255 ? ((v < 0) ? 0 : 255) : v;
                    v = p[x+2];
                    v += b / 3 / scale;
                    p[x+2] = v & ~255 ? ((v < 0) ? 0 : 255) : v;
                }
            }
        }
        if (vmode & VIDEO_CHROMA_NOISE) {
            int a, j;

            for (j = a = i = x = 0; x < pixels->pitch; x++) {
                int v = p[x] << 4;
                v += l2[j + 4] + 8;
                p[x] = (v <= 0) ? 0 : (v >= (255 << 4)) ? 255 : (v >> 4);
                a += 600 % 191;
                if (a >= 191) {
                    a -= 191;
                    i += (600 / 191) + 1;
                } else i += 600 / 191;
                while (i >= scale) {
                    i -= scale; 
                    j++;
                }
            }
        }
    }
    free(offs);
    free(ol);
    free(l);
    return pixels;
}

static tvrender_image *create_ntsc(const view64_screen *rscreen, int scale, int frame, unsigned int vmode) {
    tvrender_image *pixels;
    const static int luma[16] = {
        2, 30, 10, 20, 12, 16, 8, 24, 12, 8, 16, 10, 15, 24, 15, 20,
    };
    const static signed char cols[16][16] = {
        { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 }, // 0
        { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 }, // 1
        { -25, -5, 15, 34, 47, 53, 51, 42, 25, 5, -15, -34, -47, -53, -51, -42 }, // 2
        { 25, 7, -12, -30, -43, -49, -48, -40, -25, -7, 12, 30, 43, 49, 48, 40 }, // 3
        { 30, 48, 58, 59, 52, 36, 14, -8, -30, -48, -58, -59, -52, -36, -14, 8 }, // 4
        { -17, -36, -50, -56, -53, -43, -26, -4, 17, 36, 50, 56, 53, 43, 26, 4 }, // 5
        { 60, 55, 42, 23, 0, -23, -42, -55, -60, -55, -42, -23, 0, 23, 42, 55 }, // 6
        { -61, -57, -43, -23, 0, 23, 43, 57, 61, 57, 43, 23, 0, -23, -43, -57 }, // 7
        { -41, -21, 1, 23, 42, 55, 59, 54, 41, 21, -1, -23, -42, -55, -59, -54 }, // 8
        { -56, -41, -20, 4, 27, 47, 59, 62, 56, 41, 20, -4, -27, -47, -59, -62 }, // 9
        { -25, -5, 15, 34, 47, 53, 51, 42, 25, 5, -15, -34, -47, -53, -51, -42 }, // 10
        { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 }, // 11
        { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 }, // 12
        { -17, -36, -50, -56, -53, -43, -26, -4, 17, 36, 50, 56, 53, 43, 26, 4 }, // 13
        { 60, 55, 42, 23, 0, -23, -42, -55, -60, -55, -42, -23, 0, 23, 42, 55 }, // 14
        { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 },  // 15
    };
    unsigned char *p, *d;
    int x, y, w, lb;
    signed char *l;
    unsigned int *offs = NULL;
    d = rscreen->d[frame];
    pixels = new_tv_pixels(rscreen->w * scale, rscreen->h, (rscreen->w + 2) * scale);
    w = (pixels->pitch * 1000) / scale / 2250;
    lb = ((w * scale - pixels->w) / scale + 1) / 2;

    l = (signed char *)malloc((pixels->pitch + 3 * scale) * 784 / 253 / scale * 7 + 8 + 16);
    if (l == NULL) return NULL;
    memset(l, 0, (pixels->pitch + 3 * scale) * 784 / 253 / scale * 7 + 8 + 16);

    if (!(vmode & VIDEO_MONOCHROME)) {
        int a, i2, j;
        int scale2 = scale * 8;
        offs = (unsigned int *)malloc(sizeof(unsigned int *) * ((pixels->pitch / 3) + 2 * scale));
        if (offs == NULL) {
            free(l);
            return NULL;
        }
        for (j = i2 = a = x = 0; x < (pixels->pitch / 3) + 2 * scale; x++) {
            offs[x] = j;
            a += (784 * 3) % 253;
            if (a >= 253) {
                a -= 253;
                i2 += (784 * 3) / 253 + 1;
            } else i2 += (784 * 3) / 253;
            while (i2 >= scale2) {
                i2 -= scale2;
                j += 8;
            }
        }
    }

    memset(pixels->d, 0, pixels->pitch * pixels->h);
    for (y = 0; y < rscreen->h; y++) {
        int i, n = 0, ci = 0;
        int accu = 0;
        signed char *l2 = l;
        p = pixels->d + y * pixels->pitch;
        {
            int l;
            const signed char *c;
            int cb = rscreen->b ? rscreen->b[y * 2] : rscreen->br;
            c = cols[cb];
            for (l = 0; l < 8; l++) {
                l2[ci] = c[ci];
                ci++;
            }
        }
        for (x = -lb; x < w - lb; x++) {
            int l, ll;
            const signed char *c;
            if (x < 0) {
                int cb = rscreen->b ? rscreen->b[y * 2] : rscreen->br;
                l = luma[cb];
                c = cols[cb];
            } else if (x >= rscreen->w) {
                int cb = rscreen->b ? rscreen->b[y * 2 + 1] : rscreen->br;
                l = luma[cb];
                c = cols[cb];
            } else {
                l = luma[*d];
                c = cols[*d];
                d++;
            }
            accu += (1000 - n) * l;
            *p = accu / 128;
            p++;
            n += 2250 * scale - 1000;
            ll = 1000 * l / 128;
            while (n >= 1000) {
                *p++ = ll;
                n -= 1000;
            }
            accu = n * l;
            if (c[0]) {
                for (l = 0; l < 7; l++) {
                    l2[ci] = c[ci & 15];
                    ci++;
                }
            } else {
                memset(l2 + ci, 0, 7);
                ci += 7;
            }
        }
        p = pixels->d + y * pixels->pitch;
        if (scale == 1) {
            int a = 0, b;
            for (x = 0; x < 5; x++) a += p[x];
            b = p[x];
            p[x] = a;
            for (x = 1; x < pixels->pitch - 4; x++) {
                a += p[x+4] - b;
                b = p[x];
                p[x] = a / 5;
            }
        } else {
            int a = 0, b;
            for (x = 0; x < 7; x++) a += p[x];
            b = p[x];
            p[x] = a;
            for (x = 1; x < pixels->pitch - 6; x++) {
                a += p[x+6] - b;
                b = p[x];
                p[x] = a / 7;
            }
        }
        if (!(vmode & VIDEO_MONOCHROME)) {
            unsigned int *offs1 = offs, *offs2 = offs;
            int r = 0, g = 0, b = 0;
            for (x = 0; x < 2 * scale; x++) {
                int j = *offs2++;
                int e = l[j + 4];
                if (j & 8) {
                    r -= e;
                    g += e;
                    b += l[j + 8];
                } else {
                    r += e;
                    g -= e;
                    b -= l[j + 8];
                }
            }
            for (x = 0; x < pixels->pitch; x += 3) {
                int j;
                int v;
                int j2 = *offs1++;
                int e = l[j2 + 4];
                if (j2 & 8) {
                    r += e;
                    g -= e;
                    b -= l[j2 + 8];
                } else {
                    r -= e;
                    g += e;
                    b += l[j2 + 8];
                }
                j = *offs2++;
                e = l[j + 4];
                if (j & 8) {
                    r -= e;
                    g += e;
                    b += l[j + 8];
                } else {
                    r += e;
                    g -= e;
                    b -= l[j + 8];
                }
                v = p[x];
                v += r / 2 / scale;
                p[x] = v & ~255 ? ((v < 0) ? 0 : 255) : v;
                v = p[x+1];
                v += g * 2 / 5 / scale;
                p[x+1] = v & ~255 ? ((v < 0) ? 0 : 255) : v;
                v = p[x+2];
                v += b * 2 / 3 / scale;
                p[x+2] = v & ~255 ? ((v < 0) ? 0 : 255) : v;
            }
        }
        if (vmode & VIDEO_CHROMA_NOISE) {
            int a, j;

            for (j = a = i = x = 0; x < pixels->pitch; x++) {
                int v = p[x] << 4;
                v += l[j + 4] + 8;
                p[x] = (v <= 0) ? 0 : (v >= (255 << 4)) ? 255 : (v >> 4);
                a += 784 % 253;
                if (a >= 253) {
                    a -= 253;
                    i += (784 / 253) + 1;
                } else i += 784 / 253;
                while (i >= scale) {
                    i -= scale; j++;
                }
            }
        }
    }
    free(offs);
    free(l);
    return pixels;
}

static tvrender_image *create_rgb(const view64_screen *rscreen, int scale, int frame, const rgb_color *palette, unsigned int vmode) {
    tvrender_image *pixels;
    unsigned char *p, *d;
    int x, y, n;
    d = rscreen->d[frame];
    pixels = new_tv_pixels(rscreen->w * scale, rscreen->h, rscreen->w * scale);
    for (y = 0; y < rscreen->h; y++) {
        p = pixels->d + y * pixels->pitch;
        if (vmode & VIDEO_MONOCHROME) {
            for (x = 0; x < rscreen->w; x++) {
                memset(p, (palette[*d].r * 114 + palette[*d].g * 142) / 256, scale * 3);
                p += scale * 3;
                d++;
            }
        } else {
            for (x = 0; x < rscreen->w; x++) {
                p[0] = palette[*d].r;
                p[1] = palette[*d].g;
                p[2] = palette[*d].b;
                for (n = 1; n < scale; n++) {
                    p[3] = p[0];
                    p[4] = p[1];
                    p[5] = p[2];
                    p += 3;
                }
                p += 3;
                d++;
            }
        }
    }
    return pixels;
}

static inline unsigned char intensity_mix(unsigned char a, unsigned char b) {
    return sqrt((a * a + b * b) / 2);
}

static double intensity(double a) {
    if (a <= -1.0) return 0.0;
    if (a >= 1.0) return M_PI;
    return sqrt(1.0 - (a * a)) * a + asin(a) + (M_PI / 2.0);
}

int create_tvpicture(const view64_screen *rscreen, unsigned int vmode, int scale, const rgb_color *palette, tvrender_image *pic) {
    int x, y;
    unsigned char *p, *p2;
    tvrender_image *pixels, *pixels2;

    switch (vmode & VIDEO_MASK) {
    case VIDEO_NTSC: pixels = create_ntsc(rscreen, scale, 0, vmode); break;
    case VIDEO_PAL: pixels = create_pal(rscreen, scale, 0, vmode); break;
    default: pixels = create_rgb(rscreen, scale, 0, palette ? palette : default_palette, vmode); break;
    }
    if (pixels == NULL) return 1;
    if (rscreen->n == 2) { /* interlace */
        int i;
        switch (vmode & VIDEO_MASK) {
        case VIDEO_NTSC: pixels2 = create_ntsc(rscreen, scale, 1, vmode); break;
        case VIDEO_PAL: pixels2 = create_pal(rscreen, scale, 1, vmode); break;
        default: pixels2 = create_rgb(rscreen, scale, 1, palette ? palette : default_palette, vmode); break;
        }
        if (pixels2 == NULL) {
            free_tv_pixels(pixels);
            return 1;
        }
        p = pixels->d;
        p2 = pixels2->d;
        if (vmode & VIDEO_NOLACE && (vmode & VIDEO_MASK) == VIDEO_RGB) {
            for (y = 0; y < pixels->h; y++) {
                for (x = 0; x < pixels->pitch; x += 6 * scale) {
                    memcpy(p + x, p2 + x, 3 * scale);
                }
                p += pixels->pitch;
                p2 += pixels2->pitch;
            }
        } else {
            static unsigned short indexes[256];
            static unsigned char cache[32896];
            if (indexes[1] == 0) {
                for (x = i = 0; i < 256; i++, x += i) indexes[i] = x;
            }
            for (i = 0; i < pixels->pitch * pixels->h; i++) {
                int k = p[i] >= p2[i] ? indexes[p[i]] + p2[i] : indexes[p2[i]] + p[i];
                if (cache[k] == 0 && k > 1) {
                    cache[k] = intensity_mix(p[i], p2[i]);
                }
                p[i] = cache[k];
            }
        }
        free_tv_pixels(pixels2);
    }

    if (vmode & VIDEO_SCANLINES) { /* render scanlines */
        int s, i;
        unsigned char *e, *f;

        e = (unsigned char *)malloc(256 * (scale + 1));
        if (e == NULL) {
            free_tv_pixels(pixels);
            return 1;
        }

        for (s = 0; s <= scale; s++) {
            f = e + s * 256;
            if ((vmode & VIDEO_MASK) == VIDEO_NTSC) {
                for (x = 255; x >= 0; x--) {
                    i = x * 0.315 * (intensity((s + 0.5) / scale * (2.4 - x / 450.0)) - intensity((s - 0.5) / scale * (2.4 - x / 450.0))) * scale;
                    if (i <= 0) {
                        memset(f, 0, x + 1);
                        break;
                    }
                    f[x] = (i > 255) ? 255 : i;
                }
            } else {
                for (x = 255; x >= 0 ; x--) {
                    i = x * 0.315 * (intensity((s + 0.5) / scale * (2.0 - x / 1400.0)) - intensity((s - 0.5) / scale * (2.0 - x / 1400.0))) * scale;
                    if (i <= 0) {
                        memset(f, 0, x + 1);
                        break;
                    }
                    f[x] = (i > 255) ? 255 : i;
                }
            }
        }

        p2 = pixels->d + (pixels->pitch / 3 - pixels->w) / 2 * 3;
        memset(pic->d, 0, pic->h * pic->pitch);
        for (y = 0; y < rscreen->h; y++) {
            for (s = -scale; s <= scale; s++) {
                int y2 = y * scale + (scale / 2) + s;
                if (y2 < 0) continue;
                if (y2 >= pic->h) continue;
                f = e + abs(s) * 256;
                if (f[255] == 0) continue;
                p = pic->d + pic->pitch * y2;
                for (x = 0; x < rscreen->w * 3 * scale; x++) {
                    int v;
                    v = p[x] + f[p2[x]];
                    p[x] = v > 255 ? 255 : v;
                }
            }
            p2 += pixels->pitch;
        }
        free(e);
    } else { /* no scanlines */
        int s;
        p = pic->d;
        p2 = pixels->d + (pixels->pitch / 3 - pixels->w) / 2 * 3;
        for (y = 0; y < rscreen->h; y++) {
            for (s = 0; s < scale; s++) {
                memcpy(p, p2, rscreen->w * 3 * scale);
                p += pic->pitch;
            }
            p2 += pixels->pitch;
        }
    }
    if (vmode & VIDEO_SHADOWMASK) { /* render shadow mask */
        static unsigned char c78[256];
        static unsigned char c98[256];
        if (c78[255] == 0) {
            for (y = 0; y < 256; y++) {
                c78[y] = (y * 7) >> 3;
                c98[y] = ((y * 9) >> 3) > 255 ? 255 : ((y * 9) >> 3);
            }
        }

        for (y = 0; y < pic->h; y++) {
            p = pic->d + y * pic->pitch;
            for (x = 0; x < pic->w / 3; x++) {
                if ((y & 3) == (x & 1) * 2) {
                    p[0] = c78[p[0]];
                    p[4] = c78[p[4]];
                    p[8] = c78[p[8]];
                } else {
                    p[0] = c98[p[0]];
                    p[4] = c98[p[4]];
                    p[8] = c98[p[8]];
                }
                p[1] = c78[p[1]];
                p[2] = c78[p[2]];
                p[3] = c78[p[3]];
                p[5] = c78[p[5]];
                p[6] = c78[p[6]];
                p[7] = c78[p[7]];
                p += 9;
            }
        }
    }
    free_tv_pixels(pixels);
    return 0;
}

rgb_color *load_palette(FILE *f) {
    unsigned char values[4];
    int line = 0;
    int i, j, e = 0, c = '\n';
    rgb_color *palette;

    palette = (rgb_color *)calloc(16, sizeof(rgb_color));
    if (palette == NULL) {
        fprintf(stderr, "Not enough memory for palette\n");
        return NULL;
    }
    if (f == NULL) {
        memcpy(palette, default_palette, 16 * sizeof(rgb_color));
        return palette;
    }

    while (c != EOF && e < 16) {
        line++;
        memset(values, 0, sizeof(values));
        for (i = 0; i < 4; i++) {
            int v = 0;

            do {
                c = fgetc(f);
            } while (c == ' ' || c == '\t');

            if (c == '#' || c == '\n' || c == EOF) {
                while (c != '\n' && c != EOF) {
                    c = fgetc(f);
                }
                break;
            }

            while (c == '0') {
                c = fgetc(f);
            }

            for (j = 0; j < (2 + (i < 3)); j++) {
                if (c >= '0' && c <= '9') {
                    v = v * 16 + c - '0';
                } else if ((c | 0x20) >= 'a' && (c | 0x20) <= 'f') {
                    v = v * 16 + (c | 0x20) - 'a' + 10;
                } else {
                    break;
                }
                c = fgetc(f);
            }
            values[i] = v;
            if (j == 2 + (i < 3)) {
                fprintf(stderr, "%d: Value out of range\n", line);
            }
            while (c != ' ' && c != '\t' && c != '\n' && c != EOF) {
                c = fgetc(f);
            }
        }
        if (i > 0 && i < 4) {
            fprintf(stderr, "%d: Not enough entries\n", line);
        }
        if (c != '#' && c != '\n' && c != EOF) {
            fprintf(stderr, "%d: Junk at end of line\n", line);
        }
        if (i == 4) {
            palette[e].r = values[0];
            palette[e].g = values[1];
            palette[e].b = values[2];
            e++;
        }
    }
    if (e < 16) {
        fprintf(stderr, "Not enough colors\n");
    }
    return palette;
}
